==> diffusion-3.0.3.tar.gz: diffusion-3.0.3/main/drivers/WINSng2.cc <==
//
// WINSng2.cc    : WINSng2 Radio Support for Diffusion
// Authors       : Fabio Silva
//
// Copyright (C) 2000-2001 by the Unversity of Southern California
// $Id: WINSng2.cc,v 1.2 2001/07/12 02:01:09 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <stdio.h>
#include <stdlib.h>

#include "WINSng2.hh"

WINSNG2::WINSNG2() : DiffusionIO()
{
  int fd1, fd2;

  // Open RF radio 1
  if ((fd1 = open(RADIO_DEVICE_FILE_1, O_RDWR)) < 0){
    fprintf(stderr, "Cannot open " RADIO_DEVICE_FILE_1 " !\n");
    exit(-1);
  }

  // Open RF radio 2
  if ((fd2 = open(RADIO_DEVICE_FILE_2, O_RDWR)) < 0){
    fprintf(stderr, "Cannot open " RADIO_DEVICE_FILE_2 " !\n");
    exit(-1);
  }

  in_fds.push_back(fd1);
  out_fds.push_back(fd1);
  in_fds.push_back(fd2);
  out_fds.push_back(fd2);

  if (fd1 > fd2)
    max_in_descriptor = fd1;
  else
    max_in_descriptor = fd2;
}

Packet WINSNG2::RecvPacket(int fd)
{
  uint8 sensoria_packet_type;
  uint8 radioId[RADIO_ID_LEN];
  uint16 packet_size;
  int32_t diffId;
  Packet sensoria_packet = NULL;
  Packet incoming_packet = NULL;
  int flag;
  char *sensoria_payload;
  int sensoria_packet_size;
  struct wins_header *wins_hdr;
  struct hdr_diff *dfh;

  // Determine packet size
  flag = ioctl(fd, Ctrl_QueryFrontSize, &sensoria_packet_size);
  if (flag != 0){
    fprintf(stderr, "Cannot determine packet size ! Error Code = %d\n", flag);
    exit(-1);
  }

  sensoria_packet_size = ntohl(sensoria_packet_size);

  if (sensoria_packet_size <= 0){
    fprintf(stderr, "Incoming packet has size <= 0 !\n");
    return NULL;
  }

  // Allocate memory for the incoming packet
  sensoria_packet = new int [(sensoria_packet_size / sizeof(int)) + 1];

  if (sensoria_packet == NULL){
    fprintf(stderr, "Cannot allocate memory for the incoming radio packet !\n");
    exit(-1);
  }

  // Read the entire Sensoria Packet
  read(fd, sensoria_packet, sensoria_packet_size);
  wins_hdr = (struct wins_header *) sensoria_packet;
  sensoria_payload = ((char *) sensoria_packet) + sizeof(struct wins_header);

  // Check if packet is a diffusion packet
  sensoria_packet_type = wins_hdr->type;
  if (sensoria_packet_type != DIFFUSION_PACKET_TYPE){
    fprintf(stderr, "Received a packet type = %d, discarding...\n", sensoria_packet_type);
    delete [] sensoria_packet;
    return NULL;
  }

  // Allocate memory for the diffusion packet
  packet_size = (uint16) wins_hdr->length;
  incoming_packet = new int [(packet_size / sizeof(int)) + 1];

  if (incoming_packet == NULL){
    fprintf(stderr, "Cannot allocate memory for incoming packet !\n");
    exit(-1);
  }

  // Copy diffusion packet out of the sensoria packet
  memcpy(incoming_packet, sensoria_payload, packet_size);

  // Extract Sensoria's radio ID and diffusion's ID from packet
  dfh = HDR_DIFF(incoming_packet);
  diffId = ntohl(LAST_HOP(dfh));
  memcpy(&(radioId[0]), &(wins_hdr->destination[0]), RADIO_ID_LEN);

  // Add radio Id to Cache
  AddCache(fd, diffId, &(radioId[0]));

  // We don't need the sensoria packet anymore
  delete sensoria_packet;

  return incoming_packet;
}

void WINSNG2::SendPacket(Packet pkt, int len, int dst)
{
  uint8 radioId[RADIO_ID_LEN];
  int fd;
  char *outgoingPacket;
  list<int>::iterator itr;

  if (dst != BROADCAST_ADDR){
    fd = FindInCache(dst, &radioId[0]);
    if (fd < 0){
      fprintf(stderr, "Warning: Destination not in cache !\n");
      return;
    }

    outgoingPacket = AssemblePacket(&radioId[0], len, pkt);
    write(fd, outgoingPacket, (len + sizeof(struct wins_header)));
  }
  else{
    // We have a broadcast message
    radioId[0] = WINS_BROADCAST;
    radioId[1] = WINS_BROADCAST;
    radioId[2] = WINS_BROADCAST;

    outgoingPacket = AssemblePacket(&radioId[0], len, pkt);
    for (itr = out_fds.begin(); itr != out_fds.end(); ++itr){
      write(*itr, outgoingPacket, (len + sizeof(struct wins_header)));
    }
  }
  delete [] outgoingPacket;
}

void WINSNG2::AddCache(int fd, int32_t diffId, uint8 *srcId)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache.begin(); itr != RfCache.end(); ++itr){
    entry = *itr;
    if (memcmp(srcId, &(entry->radioId[0]), RADIO_ID_LEN) == 0){
      if ((fd == entry->radiofd) &&
	  (diffId == entry->diffusionId)){
	return;
      }
      // Something changed for this WINS address
      entry->radiofd = fd;
      entry->diffusionId = diffId;
      return;
    }
  }

  // Create a new Cache entry
  entry = new RFIdCacheEntry;
  entry->radiofd = fd;
  entry->diffusionId = diffId;
  memcpy(&(entry->radioId[0]), srcId, RADIO_ID_LEN);
  RfCache.push_back(entry);
}

int WINSNG2::FindInCache(int32_t diffId, uint8 *srcId)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache.begin(); itr != RfCache.end(); ++itr){
    entry = *itr;

    if (diffId == entry->diffusionId){
      // Entry found !
      
      memcpy(srcId, &(entry->radioId[0]), RADIO_ID_LEN);
      return entry->radiofd;
    }
  }

  // Entry not in the cache
  return -1;
}

char * WINSNG2::AssemblePacket(uint8 *srcId, int len, Packet pkt)
{
  char *outgoingPacket;
  char *payload;
  struct wins_header *wins_hdr;

  // Allocate memory for packet + header
  outgoingPacket = new char[len + sizeof(struct wins_header)];
  wins_hdr = (struct wins_header *) outgoingPacket;
  payload = outgoingPacket + sizeof(struct wins_header);

  // Fill wins header  
  wins_hdr->length = (uint16) len;
  wins_hdr->source[0] = 0;
  wins_hdr->source[1] = 0;
  wins_hdr->source[2] = 0;
  memcpy(&(wins_hdr->destination[0]), srcId, RADIO_ID_LEN);
  wins_hdr->type = (uint8) DIFFUSION_PACKET_TYPE;

  // Copy packet
  memcpy(payload, pkt, len);

  return outgoingPacket;
}

==> diffusion-3.0.4.tar.gz: diffusion-3.0.4/main/drivers/WINSng2.cc <==
//
// WINSng2.cc    : WINSng2 Radio Support for Diffusion
// Authors       : Fabio Silva
//
// Copyright (C) 2000-2001 by the Unversity of Southern California
// $Id: WINSng2.cc,v 1.3 2001/09/07 23:10:30 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <stdio.h>
#include <stdlib.h>

#include "WINSng2.hh"

WINSNG2::WINSNG2() : DiffusionIO()
{
  int fd1, fd2;

  // Open RF radio 1
  if ((fd1 = open(RADIO_DEVICE_FILE_1, O_RDWR)) < 0){
    fprintf(stderr, "Cannot open " RADIO_DEVICE_FILE_1 " !\n");
    exit(-1);
  }

  // Open RF radio 2
  if ((fd2 = open(RADIO_DEVICE_FILE_2, O_RDWR)) < 0){
    fprintf(stderr, "Cannot open " RADIO_DEVICE_FILE_2 " !\n");
    exit(-1);
  }

  in_fds.push_back(fd1);
  out_fds.push_back(fd1);
  in_fds.push_back(fd2);
  out_fds.push_back(fd2);

  if (fd1 > fd2)
    max_in_descriptor = fd1;
  else
    max_in_descriptor = fd2;
}

Packet WINSNG2::RecvPacket(int fd)
{
  uint8 sensoria_packet_type;
  uint8 radioId[RADIO_ID_LEN];
  uint16 packet_size;
  int32_t diffId;
  Packet sensoria_packet = NULL;
  Packet incoming_packet = NULL;
  int flag;
  char *sensoria_payload;
  int sensoria_packet_size;
  struct wins_header *wins_hdr;
  struct hdr_diff *dfh;

  // Determine packet size
  flag = ioctl(fd, Ctrl_QueryFrontSize, &sensoria_packet_size);
  if (flag != 0){
    fprintf(stderr, "Cannot determine packet size ! Error Code = %d\n", flag);
    exit(-1);
  }

  sensoria_packet_size = ntohl(sensoria_packet_size);

  if (sensoria_packet_size <= 0){
    fprintf(stderr, "Incoming packet has size <= 0 !\n");
    return NULL;
  }

  // Allocate memory for the incoming packet
  sensoria_packet = new int [(sensoria_packet_size / sizeof(int)) + 1];

  if (sensoria_packet == NULL){
    fprintf(stderr, "Cannot allocate memory for the incoming radio packet !\n");
    exit(-1);
  }

  // Read the entire Sensoria Packet
  read(fd, sensoria_packet, sensoria_packet_size);
  wins_hdr = (struct wins_header *) sensoria_packet;
  sensoria_payload = ((char *) sensoria_packet) + sizeof(struct wins_header);

  // Check if packet is a diffusion packet
  sensoria_packet_type = wins_hdr->type;
  if (sensoria_packet_type != DIFFUSION_PACKET_TYPE){
    fprintf(stderr, "Received a packet type = %d, discarding...\n", sensoria_packet_type);
    delete [] sensoria_packet;
    return NULL;
  }

  // Allocate memory for the diffusion packet
  packet_size = (uint16) wins_hdr->length;
  incoming_packet = new int [(packet_size / sizeof(int)) + 1];

  if (incoming_packet == NULL){
    fprintf(stderr, "Cannot allocate memory for incoming packet !\n");
    exit(-1);
  }

  // Copy diffusion packet out of the sensoria packet
  memcpy(incoming_packet, sensoria_payload, packet_size);

  // Extract Sensoria's radio ID and diffusion's ID from packet
  dfh = HDR_DIFF(incoming_packet);
  diffId = ntohl(LAST_HOP(dfh));
  memcpy(&(radioId[0]), &(wins_hdr->source[0]), RADIO_ID_LEN);

  // Add radio Id to Cache
  AddCache(fd, diffId, &(radioId[0]));

  // We don't need the sensoria packet anymore
  delete sensoria_packet;

  return incoming_packet;
}

void WINSNG2::SendPacket(Packet pkt, int len, int dst)
{
  uint8 radioId[RADIO_ID_LEN];
  int fd;
  char *outgoingPacket;
  list<int>::iterator itr;

  if (dst != BROADCAST_ADDR){
    fd = FindInCache(dst, &radioId[0]);
    if (fd < 0){
      fprintf(stderr, "Warning: Destination not in cache !\n");
      return;
    }

    outgoingPacket = AssemblePacket(&radioId[0], len, pkt);
    write(fd, outgoingPacket, (len + sizeof(struct wins_header)));
  }
  else{
    // We have a broadcast message
    radioId[0] = WINS_BROADCAST;
    radioId[1] = WINS_BROADCAST;
    radioId[2] = WINS_BROADCAST;

    outgoingPacket = AssemblePacket(&radioId[0], len, pkt);
    for (itr = out_fds.begin(); itr != out_fds.end(); ++itr){
      write(*itr, outgoingPacket, (len + sizeof(struct wins_header)));
    }
  }
  delete [] outgoingPacket;
}

void WINSNG2::AddCache(int fd, int32_t diffId, uint8 *srcId)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache.begin(); itr != RfCache.end(); ++itr){
    entry = *itr;
    if (memcmp(srcId, &(entry->radioId[0]), RADIO_ID_LEN) == 0){
      if ((fd == entry->radiofd) &&
	  (diffId == entry->diffusionId)){
	return;
      }
      // Something changed for this WINS address
      entry->radiofd = fd;
      entry->diffusionId = diffId;
      return;
    }
  }

  // Create a new Cache entry
  entry = new RFIdCacheEntry;
  entry->radiofd = fd;
  entry->diffusionId = diffId;
  memcpy(&(entry->radioId[0]), srcId, RADIO_ID_LEN);
  RfCache.push_back(entry);
}

int WINSNG2::FindInCache(int32_t diffId, uint8 *srcId)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache.begin(); itr != RfCache.end(); ++itr){
    entry = *itr;

    if (diffId == entry->diffusionId){
      // Entry found !
      
      memcpy(srcId, &(entry->radioId[0]), RADIO_ID_LEN);
      return entry->radiofd;
    }
  }

  // Entry not in the cache
  return -1;
}

char * WINSNG2::AssemblePacket(uint8 *srcId, int len, Packet pkt)
{
  char *outgoingPacket;
  char *payload;
  struct wins_header *wins_hdr;

  // Allocate memory for packet + header
  outgoingPacket = new char[len + sizeof(struct wins_header)];
  wins_hdr = (struct wins_header *) outgoingPacket;
  payload = outgoingPacket + sizeof(struct wins_header);

  // Fill wins header  
  wins_hdr->length = (uint16) len;
  wins_hdr->source[0] = 0;
  wins_hdr->source[1] = 0;
  wins_hdr->source[2] = 0;
  memcpy(&(wins_hdr->destination[0]), srcId, RADIO_ID_LEN);
  wins_hdr->type = (uint8) DIFFUSION_PACKET_TYPE;

  // Copy packet
  memcpy(payload, pkt, len);

  return outgoingPacket;
}
